//****************************************************************************//
//****************************************************************************//
//******     LZB******************//
//******   ,    *********//
//******  ,    .*************************//
//****************************************************************************//
//****************************************************************************//


unit LZSS;


interface

uses Windows, Classes, SysUtils;

const
  THRESHOLD   = 3;
type
  TLZ = Class
  public
    function Compress(const Source; var Dest; Size: ULONG; Without_Size: Boolean = false; BigEndian: Boolean = true): ULONG;
    function Decompress(const Source; var Dest; Without_Size: Boolean = false; BigEndian: Boolean = true): ULONG;
    constructor Create;
  private
    DataStreamPos: ULONG;
    CompressedStreamPos: ULONG;
    Data:array of UCHAR;
    CompressedData:array of UCHAR;
    MatchPos, MatchLen: ULONG;
    DataSize: ULONG;
    CompressedSize: ULONG;
    CompressedLong: ULONG;
    CompressedBitsUsed: UINT;

    //Functions

    function  FindMatches(): Boolean;
    procedure CompressedStreamReadBits(Value: PUINT; NumBits: UINT);overload;
    function  CompressedStreamReadBits(NumBits: UINT): UINT;overload;
    function  CompressedStreamReadBits: Boolean;overload;
    procedure CompressedStreamWriteBits(Value : UINT; NumBits: UINT); overload;
    procedure CompressedStreamWriteBits(Value : Byte); overload;
    procedure CompressedStreamWriteBits(Value : Boolean); overload;
    procedure CompressedStreamWriteBitsFlush();
    function  DecodeLength: Integer;
    procedure EncodeLength(Length: Integer);
    function  DecodeOffset: Integer;
    procedure EncodeOffset();
end;

var
  //,     .
  p_len: array[0..255] of byte = (
	$00, $01, $02, $02, $03, $03, $03, $03, $04, $04, $04, $04, $04, $04, $04, $04, 
	$05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, 
	$06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, 
	$06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, 
	$07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, 
	$07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, 
	$07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, 
	$07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, 
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, 
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, 
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, 
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08,
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08,
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08,
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08,
	$08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08
);
implementation


procedure TLZ.CompressedStreamReadBits(Value: PUINT; NumBits: UINT);
var
  Temp: ULONG;
begin
  CompressedLong:= CompressedLong and $FFFF;
  while NumBits > 0 do
  begin
    dec(NumBits);
    if CompressedBitsUsed = 0 then
    begin
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or (Temp shl 8);
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or Temp;
      CompressedBitsUsed:= 16;
    end;
    CompressedLong:= CompressedLong shl 1;
    dec(CompressedBitsUsed);
  end;
  Value^:= CompressedLong shr 16;
end;

function TLZ.CompressedStreamReadBits(NumBits: UINT): UINT;
var
  Temp: ULONG;
begin
  CompressedLong:= CompressedLong and $FFFF;
  while NumBits > 0 do
  begin
    dec(NumBits);
    if CompressedBitsUsed = 0 then
    begin
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or (Temp shl 8);
      Temp:= CompressedData[CompressedStreamPos];
      inc(CompressedStreamPos);
      CompressedLong:= CompressedLong or Temp;
      CompressedBitsUsed:= 16;
    end;
    CompressedLong:= CompressedLong shl 1;
    dec(CompressedBitsUsed);
  end;
  Result:= CompressedLong shr 16;
end;

function TLZ.CompressedStreamReadBits: Boolean;
var
  Temp: ULONG;
begin
  CompressedLong:= CompressedLong and $FFFF;
  if CompressedBitsUsed = 0 then
  begin
    Temp:= CompressedData[CompressedStreamPos];
    inc(CompressedStreamPos);
    CompressedLong:= CompressedLong or (Temp shl 8);
    Temp:= CompressedData[CompressedStreamPos];
    inc(CompressedStreamPos);
    CompressedLong:= CompressedLong or Temp;
    CompressedBitsUsed:= 16;
  end;
  CompressedLong:= CompressedLong shl 1;
  dec(CompressedBitsUsed);
  Result:= Boolean(CompressedLong shr 16);
end;

procedure TLZ.CompressedStreamWriteBits(Value, NumBits: UINT);
begin
  while NumBits > 0 do
  begin
    dec(NumBits);
    CompressedLong := CompressedLong shl 1;
    CompressedLong:= CompressedLong or ((Value shr NumBits) and 1);
    inc(CompressedBitsUsed);

    if(CompressedBitsUsed = 8) then
    begin
      CompressedData[CompressedStreamPos]:= CompressedLong;
      inc(CompressedStreamPos);
      inc(CompressedSize);
      CompressedBitsUsed:= 0;
    end;
  end;  //end while
end;

procedure TLZ.CompressedStreamWriteBits(Value: Byte);
begin
  CompressedLong := CompressedLong shl 1;
  CompressedLong:= CompressedLong or (Value and 1);
  inc(CompressedBitsUsed);

  if(CompressedBitsUsed = 8) then
  begin
    CompressedData[CompressedStreamPos]:= CompressedLong;
    inc(CompressedStreamPos);
    inc(CompressedSize);
    CompressedBitsUsed:= 0;
  end;
end;

procedure TLZ.CompressedStreamWriteBits(Value: Boolean);
begin
  CompressedLong := CompressedLong shl 1;
  CompressedLong:= CompressedLong or Byte(Value);
  inc(CompressedBitsUsed);

  if(CompressedBitsUsed = 8) then
  begin
    CompressedData[CompressedStreamPos]:= CompressedLong;
    inc(CompressedStreamPos);
    inc(CompressedSize);
    CompressedBitsUsed:= 0;
  end;
end;


procedure TLZ.CompressedStreamWriteBitsFlush;
begin
  if CompressedBitsUsed > 0 then
  begin
      CompressedData[CompressedStreamPos]:= CompressedLong shl (8 - CompressedBitsUsed);
      inc(CompressedStreamPos);
      inc(CompressedSize);
  end;
end;


function TLZ.DecodeLength: Integer;
begin
  Result:= 1;
  while not CompressedStreamReadBits do
  begin
    Result:= Result shl 1;
    Result:= Result or CompressedStreamReadBits(1);
  end;
end;

function TLZ.DecodeOffset(): Integer;
begin
  if DataStreamPos > 255 then
  begin
    Result:= CompressedStreamReadBits(p_len[DataStreamPos shr 8] + 8);
  end else
  begin
    Result:= CompressedStreamReadBits(p_len[DataStreamPos]);
  end;
end;

//    ,   .
//    .
//      .   :
//     1      ,
//    1.   ,    .
//    :       0.      1.
//      .    15    .
//   15 = 1111
//      111
//     0   010101
//   1       0101011
//   .
procedure TLZ.EncodeLength(Length: Integer);
var
  I, J: Integer;
begin
  I:= Length;
  J:= 0;
  while I > 0 do
  begin
    I:= I shr 1;
    Inc(J);
  end;
  Dec(J);
  while J > 0 do
  begin
    Dec(J);
    CompressedStreamWriteBits(0, 1);
    CompressedStreamWriteBits(Length shr J, 1);
  end;
  CompressedStreamWriteBits(TRUE);
end;

//********************************************************************************//
//**   . ,     2,**//
//**    1 , ,  4-   ,  *****//
//** 2 ,  ..,      N .*****************//
//********************************************************************************//
procedure TLZ.EncodeOffset();
var
  I: Integer;
begin
  if DataStreamPos > 255 then
  begin
    I:= 8 + p_len[DataStreamPos shr 8];
  end else
    I:= p_len[DataStreamPos];
  CompressedStreamWriteBits(MatchPos, I);
end;

function TLZ.Compress(const Source; var Dest; Size: ULONG; Without_Size: Boolean = false; BigEndian: Boolean = true): ULONG;
var
  Buffer, Literals: array[Word] of Byte;
  LitPos: Integer;
  LitCount: Integer;
begin
  Data:= @Source;
  CompressedData:= @Dest;
  DataSize:= Size;
  LitPos:= 0;
  LitCount:= 0;
  while not FindMatches do
  begin
    Inc(LitCount);
    Inc(DataStreamPos);
  end;
  Move(Data[DataStreamPos - LitCount], Literals[LitPos], LitCount);
  Inc(LitPos, LitCount);
  EncodeLength(LitCount);
  if FindMatches then
  begin
    EncodeOffset;
    Inc(DataStreamPos, MatchLen);
    EncodeLength(MatchLen - 2);
  end;
  while DataStreamPos < Size do
  begin
//     .   .
    if FindMatches then
    begin
      CompressedStreamWriteBits(True);
      EncodeOffset;
      Inc(DataStreamPos, MatchLen);
      EncodeLength(MatchLen - 2);
    end else
    begin
      LitCount:= 0;
      while (not FindMatches) and (DataStreamPos < DataSize) do
      begin
        Inc(LitCount);
        Inc(DataStreamPos);
      end;
      CompressedStreamWriteBits(False);
      Move(Data[DataStreamPos - LitCount], Literals[LitPos], LitCount);
      Inc(LitPos, LitCount);
      EncodeLength(LitCount);
      if FindMatches then
      begin
        EncodeOffset;
        Inc(DataStreamPos, MatchLen);
        EncodeLength(MatchLen - 2);
      end
    end;

  //  
//    LitCount:= 0;
//    while not FindMatches do
//    begin
//      Inc(LitCount);
//      Inc(DataStreamPos);
//    end;
//    Move(Data[DataStreamPos - LitCount], Literals[LitPos], LitCount);
//    Inc(LitPos, LitCount);
//    EncodeLength(LitCount);
//    EncodeOffset;
//    Inc(DataStreamPos, MatchLen);
//    EncodeLength(MatchLen - 2);
//    while FindMatches do
//    begin
//      CompressedStreamWriteBit(True);
//      EncodeOffset;
//      Inc(DataStreamPos, MatchLen);
//      EncodeLength(MatchLen - 2);
//    end;
//    CompressedStreamWriteBit(False);

//      
//    if FindMatches then
//    begin
//      if not NewCycle then
//        CompressedStreamWriteBit(True);
//      EncodeOffset;
//      Inc(DataStreamPos, MatchLen);
//      EncodeLength(MatchLen - 2);
//      NewCycle:= False;
//    end else
//    begin
//      if not NewCycle then
//        CompressedStreamWriteBit(False);
//      LitCount:= 0;
//      while not FindMatches do
//      begin
//        Inc(LitCount);
//        Inc(DataStreamPos);
//        if (DataStreamPos >= DataSize) then
//          break;
//      end;
//      Move(Data[DataStreamPos - LitCount], Literals[LitPos], LitCount);
//      Inc(LitPos, LitCount);
//      EncodeLength(LitCount);
//      NewCycle:= True;
//    end;  

  end;
  CompressedStreamWriteBitsFlush;
  if Without_Size then
  begin
    if BigEndian then
    begin
      Buffer[0]:= (LitPos + 2) shr 8;
      Buffer[1]:= (LitPos + 2);
    end else
    begin
      Buffer[1]:= (LitPos + 2) shr 8;
      Buffer[0]:= (LitPos + 2);
    end;
    Move(Literals[0], Buffer[2], LitPos);
    Move(CompressedData[0], Buffer[LitPos + 2], CompressedStreamPos);
    Move(Buffer[0], CompressedData[0], Length(CompressedData));
    Result:= CompressedStreamPos + LitPos + 2;
  end else
  begin
    if BigEndian then
    begin
      Buffer[0]:= DataSize shr 8;
      Buffer[1]:= DataSize;
      Buffer[2]:= (LitPos + 2) shr 8;
      Buffer[3]:= (LitPos + 2);
    end else
    begin
      Buffer[1]:= DataSize shr 8;
      Buffer[0]:= DataSize;
      Buffer[3]:= (LitPos + 2) shr 8;
      Buffer[2]:= (LitPos + 2);
    end;
    Move(Literals[0], Buffer[4], LitPos);
    Move(CompressedData[0], Buffer[LitPos + 4], CompressedStreamPos);
    Move(Buffer[0], CompressedData[0], Length(CompressedData));
    Result:= CompressedStreamPos + LitPos + 4;
  end;
end;


function TLZ.Decompress(const Source; var Dest; Without_Size: Boolean = false; BigEndian: Boolean = True): ULONG;
var
  LiteralPos: Integer;
  Bit: Boolean;
begin
  CompressedData:= @Source;
  Data:= @Dest;
  
  if Without_Size then
  begin
    LiteralPos:= 2;
    DataSize:= Length(Data);
    if BigEndian then
      CompressedStreamPos:= CompressedStreamReadBits(16)
    else
      CompressedStreamPos:= Swap(CompressedStreamReadBits(16));
  end
  else
  begin
    LiteralPos:= 4;
    if BigEndian then
    begin
      DataSize:= CompressedStreamReadBits(16);
      CompressedStreamPos:= CompressedStreamReadBits(16) + 2;
    end else
    begin
      DataSize:= Swap(CompressedStreamReadBits(16));
      CompressedStreamPos:= Swap(CompressedStreamReadBits(16)) + 2;
    end;  

  end;

  while DataStreamPos < DataSize do
  begin
    MatchLen:= DecodeLength();
    Move(CompressedData[LiteralPos], Data[DataStreamPos], MatchLen);
    Inc(DataStreamPos, MatchLen);
    Inc(LiteralPos, MatchLen);
    if DataStreamPos >= DataSize then
      Break;
    Bit:= True;
    while Bit do
    begin
      MatchPos:= DecodeOffset();
      MatchLen:= DecodeLength() + 2;
      while MatchLen > 0 do
      begin
        Data[DataStreamPos]:= Data[MatchPos];
        Inc(DataStreamPos);
        Inc(MatchPos);
        Dec(MatchLen);
      end;
      if DataStreamPos >= DataSize then
        Break;
      Bit:= CompressedStreamReadBits;
    end;

  end;
  Result:= CompressedStreamPos;
end;

function Compare(P1, P2: Pointer; Length: Integer): Boolean; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     ESI,P1
        MOV     EDI,P2
        MOV     EDX,ECX
        XOR     EAX,EAX
        AND     EDX,3
        SAR     ECX,2
        JS      @@1     // Negative Length implies identity.
        REPE    CMPSD
        JNE     @@2
        MOV     ECX,EDX
        REPE    CMPSB
        JNE     @@2
@@1:    INC     EAX
@@2:    POP     EDI
        POP     ESI
end;

function TLZ.FindMatches: Boolean;
var
  MaxPos: LongWord;
  BufPos: LongWord;
  TempLen: Integer;
begin
  MatchLen:= THRESHOLD - 2;
  BufPos:= 0;
  MaxPos:= DataStreamPos;
  TempLen:= THRESHOLD;
  begin
    while BufPos < MaxPos do
    begin
      while (Compare(@Data[BufPos], @Data[DataStreamPos], TempLen) and ((DataStreamPos + TempLen) <= DataSize)) do
      begin
        MatchLen:= TempLen;
        MatchPos:= BufPos;
        inc(TempLen);
      end;
      inc(BufPos);
    end;
  end;
  if MatchLen < THRESHOLD then
    Result:= False
  else
    Result:= True;
end;

constructor TLZ.Create;
begin
  DataStreamPos:= 0;
  CompressedStreamPos:= 0;
  CompressedLong:= 0;
  CompressedBitsUsed:= 0;
end;

end.
